@@ -26,6 +26,7 @@ import datetime
 import shutil
 import re
 import pathlib
+import bisect
 from contextlib import contextmanager
 import matplotlib
 from PySide2.QtCore import Qt, Slot, QFile, QIODevice, QSize, QRect, QPoint, QUrl, QObject, QEvent
@@ -33,6 +34,7 @@ from PySide2.QtCore import __version__ as qt_version
 from PySide2.QtCore import __version_info__ as qt_version_info
 from PySide2.QtWidgets import QApplication, QMessageBox, QFileIconProvider, QStyle, QFileDialog, QInputDialog
 from PySide2.QtGui import (
+    QGuiApplication,
     QCursor,
     QImageReader,
     QPixmap,
@@ -42,7 +44,6 @@ from PySide2.QtGui import (
     QStandardItem,
     QDesktopServices,
     QKeySequence,
-    QTextCursor,
     QPalette,
     QSyntaxHighlighter,
     QTextCharFormat,
@@ -53,7 +54,13 @@ from PySide2.QtGui import (
 )
 from spine_engine.utils.serialization import deserialize_path
 from spinedb_api.spine_io.gdx_utils import find_gams_directory
-from .config import DEFAULT_WORK_DIR, PLUGINS_PATH
+from .config import (
+    DEFAULT_WORK_DIR,
+    PLUGINS_PATH,
+    PROJECT_FILENAME,
+    PROJECT_LOCAL_DATA_DIR_NAME,
+    PROJECT_LOCAL_DATA_FILENAME,
+)
 
 if os.name == "nt":
     import ctypes
@@ -98,25 +105,8 @@ def format_log_message(msg_type, message, show_datetime=True):
     return open_tag + date_str + message + "</span>"
 
 
-def add_message_to_document(document, message):
-    """Adds a message to a document and return the cursor.
-
-    Args:
-        document (QTextDocument)
-        message (str)
-
-    Returns:
-        QTextCursor
-    """
-    cursor = QTextCursor(document)
-    cursor.movePosition(QTextCursor.End)
-    cursor.insertBlock()
-    cursor.insertHtml(message)
-    return cursor
-
-
 def busy_effect(func):
-    """ Decorator to change the mouse cursor to 'busy' while a function is processed.
+    """Decorator to change the mouse cursor to 'busy' while a function is processed.
 
     Args:
         func (Callable): Decorated function.
@@ -415,18 +405,19 @@ def format_string_list(str_list):
 
 
 def rows_to_row_count_tuples(rows):
-    """Breaks a list of rows into a list of (row, count) tuples corresponding
-    to chunks of successive rows.
+    """Breaks a list of rows into a list of (row, count) tuples to corresponding
+    chunks of successive rows.
 
     Args:
-        rows (list): rows
+        rows (Iterable): rows
 
     Returns:
         list of tuple: row count tuples
     """
+    rows = set(rows)
     if not rows:
         return []
-    sorted_rows = sorted(set(rows))
+    sorted_rows = sorted(rows)
     break_points = [k + 1 for k in range(len(sorted_rows) - 1) if sorted_rows[k] + 1 != sorted_rows[k + 1]]
     break_points = [0] + break_points + [len(sorted_rows)]
     ranges = [(break_points[l], break_points[l + 1]) for l in range(len(break_points) - 1)]
@@ -518,7 +509,7 @@ class CharIconEngine(TransparentIconEngine):
         """
         super().__init__()
         self.char = char
-        self.color = color
+        self.color = QColor(color)
         self.font = QFont('Font Awesome 5 Free Solid')
 
     def paint(self, painter, rect, mode=None, state=None):
@@ -548,8 +539,8 @@ class ColoredIcon(QIcon):
     def set_colored(self, colored):
         self._engine.set_colored(colored)
 
-    def color(self):
-        return self._engine.color()
+    def color(self, mode=QIcon.Normal):
+        return self._engine.color(mode=mode)
 
 
 class ColoredIconEngine(QIconEngine):
@@ -558,26 +549,37 @@ class ColoredIconEngine(QIconEngine):
         self._icon = QIcon(icon_file_name)
         self._icon_color = icon_color
         self._base_pixmap = self._icon.pixmap(icon_size)
-        self._pixmap = None
         self._colored = None
+        self._pixmaps = {}
         self.set_colored(colored)
 
-    def color(self):
-        if not self._colored:
-            return None
-        return self._icon_color
+    def color(self, mode=QIcon.Normal):
+        color = self._icon_color if self._colored else QColor("black")
+        if mode == QIcon.Disabled:
+            r, g, b, a = color.getRgbF()
+            tint = 0.37255
+            color = QColor.fromRgbF(r + (1.0 - r) * tint, g + (1.0 - g) * tint, b + (1.0 - b) * tint, a)
+        return color
 
     def set_colored(self, colored):
         if self._colored == colored:
             return
         self._colored = colored
-        if colored:
-            self._pixmap = color_pixmap(self._base_pixmap, self._icon_color)
-        else:
-            self._pixmap = self._base_pixmap
+        self._pixmaps.clear()
+
+    def _do_make_pixmap(self, mode, state):
+        color = self.color(mode)
+        return color_pixmap(self._base_pixmap, color)
+
+    def _make_pixmap(self, mode, state):
+        if (mode, state) not in self._pixmaps:
+            self._pixmaps[mode, state] = self._do_make_pixmap(mode, state)
+        return self._pixmaps[mode, state]
 
     def pixmap(self, size, mode, state):
-        return self._pixmap.scaled(self._icon.actualSize(size), Qt.KeepAspectRatio, Qt.SmoothTransformation)
+        return self._make_pixmap(mode, state).scaled(
+            self._icon.actualSize(size), Qt.KeepAspectRatio, Qt.SmoothTransformation
+        )
 
 
 def color_pixmap(pixmap, color):
@@ -961,17 +963,21 @@ def make_settings_dict_for_engine(app_settings):
     Returns:
         dict: Engine-compatible settings
     """
-    # XXX: We may want to introduce a new group "executionSettings", for more clarity
+
+    def dump_group(group):
+        app_settings.beginGroup(group)
+        for key in app_settings.childKeys():
+            value = app_settings.value(key)
+            try:
+                json.dumps(value)
+            except (TypeError, json.decoder.JSONDecodeError):
+                continue
+            settings[f"{group}/{key}"] = value
+        app_settings.endGroup()
+
     settings = {}
-    app_settings.beginGroup("appSettings")
-    for key in app_settings.childKeys():
-        value = app_settings.value(key)
-        try:
-            json.dumps(value)
-        except (TypeError, json.decoder.JSONDecodeError):
-            continue
-        settings[f"appSettings/{key}"] = value
-    app_settings.endGroup()
+    dump_group("appSettings")
+    dump_group("engineSettings")
     if "appSettings/workDir" not in settings:
         # Headless mode may execute on a system where we don't have any Toolbox settings available.
         # Make sure we set a sane work directory for Tools, at least.
@@ -1027,13 +1033,13 @@ def get_upgrade_db_promt_text(url, current, expected):
     text = (
         f"The database at <b>{url}</b> is at revision <b>{current}</b> and needs to be "
         f"upgraded to revision <b>{expected}</b> in order to be used with the current "
-        f"version of Spine Toolbox."
+        "version of Spine Toolbox."
     )
     info_text = (
-        "Do you want to upgrade the database now?"
-        "<p><b>WARNING</b>: After the upgrade, "
+        "<b>WARNING</b>: After the upgrade, "
         "the database may no longer be used "
         "with previous versions of Spine."
+        "<p>Do you want to upgrade the database now?"
     )
     return text, info_text
 
@@ -1058,6 +1064,9 @@ def parse_specification_file(spec_path, logger):
     except FileNotFoundError:
         logger.msg_error.emit(f"Specification file <b>{spec_path}</b> does not exist")
         return None
+    except OSError:
+        logger.msg_error.emit(f"Specification file <b>{spec_path}</b> not found")
+        return None
 
 
 def load_specification_from_file(spec_path, spec_factories, app_settings, logger):
@@ -1181,6 +1190,27 @@ def load_plugin_specifications(plugin_dict, spec_factories, app_settings, logger
     return {name: plugin_specs}
 
 
+DB_ITEM_SEPARATOR = " \u01C0 "
+"""Display string to separate items such as entity names."""
+
+
+def parameter_identifier(database, parameter, names, alternative):
+    """Concatenates given information into parameter value identifier string.
+
+    Args:
+        database (str, optional): database's code name
+        parameter (str): parameter's name
+        names (list of str): name of the entity or class that holds the value
+        alternative (str or NoneType): name of the value's alternative
+    """
+    parts = [database] if database is not None else []
+    parts += [parameter]
+    if alternative is not None:
+        parts += [alternative]
+    parts += [DB_ITEM_SEPARATOR.join(names)]
+    return " - ".join(parts)
+
+
 class SignalWaiter(QObject):
     """A 'traffic light' that allows waiting for a signal to be emitted in another thread."""
 
@@ -1208,6 +1238,7 @@ def signal_waiter(signal):
         yield waiter
     finally:
         signal.disconnect(waiter.trigger)
+        waiter.deleteLater()
 
 
 class CustomSyntaxHighlighter(QSyntaxHighlighter):
@@ -1237,9 +1268,11 @@ class CustomSyntaxHighlighter(QSyntaxHighlighter):
         if self.lexer is None:
             return ()
         for start, ttype, subtext in self.lexer.get_tokens_unprocessed(text):
-            while ttype not in self._formats:
+            while True:
+                text_format = self._formats.get(ttype)
+                if text_format is not None:
+                    break
                 ttype = ttype.parent
-            text_format = self._formats.get(ttype, QTextCharFormat())
             yield start, len(subtext), text_format
 
     def highlightBlock(self, text):
@@ -1264,16 +1297,223 @@ def inquire_index_name(model, column, title, parent_widget):
     model.setHeaderData(column, Qt.Horizontal, new_name)
 
 
-class CacheItem(dict):
-    """A dictionary that behaves kinda like a row from a query result.
+def preferred_row_height(widget, factor=1.5):
+    return factor * widget.fontMetrics().lineSpacing()
 
-    It is used to store items in a cache, so we can access them as if they were rows from a query result.
-    This is mainly because we want to use the cache as a replacement for db queries in some methods.
+
+def restore_ui(window, app_settings, settings_group):
+    """Restores UI state from previous session.
+
+    Args:
+        window (QMainWindow)
+        app_settings (QSettings)
+        settings_group (str)
     """
+    app_settings.beginGroup(settings_group)
+    window_size = app_settings.value("windowSize")
+    window_pos = app_settings.value("windowPosition")
+    window_state = app_settings.value("windowState")
+    window_maximized = app_settings.value("windowMaximized", defaultValue='false')
+    n_screens = app_settings.value("n_screens", defaultValue=1)
+    app_settings.endGroup()
+    original_size = window.size()
+    if window_size:
+        window.resize(window_size)
+    if window_pos:
+        window.move(window_pos)
+    if window_state:
+        window.restoreState(window_state, version=1)  # Toolbar and dockWidget positions
+    # noinspection PyArgumentList
+    if len(QGuiApplication.screens()) < int(n_screens):
+        # There are less screens available now than on previous application startup
+        window.move(0, 0)  # Move this widget to primary screen position (0,0)
+    ensure_window_is_on_screen(window, original_size)
+    if window_maximized == 'true':
+        window.setWindowState(Qt.WindowMaximized)
 
-    def __getattr__(self, name):
-        """Overridden method to return the dictionary key named after the attribute, or None if it doesn't exist."""
-        return self.get(name)
 
-    def _asdict(self):
-        return dict(**self)
+def save_ui(window, app_settings, settings_group):
+    """Saves UI state for next session.
+
+    Args:
+        window (QMainWindow)
+        app_settings (QSettings)
+        settings_group (str)
+    """
+    app_settings.beginGroup(settings_group)
+    app_settings.setValue("windowSize", window.size())
+    app_settings.setValue("windowPosition", window.pos())
+    app_settings.setValue("windowState", window.saveState(version=1))
+    app_settings.setValue("windowMaximized", window.windowState() == Qt.WindowMaximized)
+    app_settings.setValue("n_screens", len(QGuiApplication.screens()))
+    app_settings.endGroup()
+
+
+def bisect_chunks(current_data, new_data, key=None):
+    if key is not None:
+        current_data = [key(x) for x in current_data]
+    else:
+        key = lambda x: x
+    new_data = sorted(new_data, key=key)
+    if not new_data:
+        return ()
+    item = new_data[0]
+    chunk = [item]
+    lo = bisect.bisect_left(current_data, key(item))
+    for item in new_data[1:]:
+        row = bisect.bisect_left(current_data, key(item), lo=lo)
+        if row == lo:
+            chunk.append(item)
+            continue
+        yield chunk, lo
+        count = len(chunk)
+        chunk = [item]
+        lo = row + count
+    yield chunk, lo
+
+
+class FetchParent:
+    @property
+    def fetch_item_type(self):
+        """Returns the type of item to fetch, e.g., "object_class".
+        Used to create an initial query for this item.
+
+        Returns:
+            str
+        """
+        raise NotImplementedError()
+
+    # pylint: disable=no-self-use
+    def filter_query(self, query, subquery, db_map):
+        """Filters the initial query created using the ``fetch_item_type`` property.
+
+        Args:
+            query (Query): The query
+            subquery (Alias): The source of the query
+            db_map (DiffDatabaseMapping)
+
+        Returns:
+            Query
+        """
+        return query
+
+    def fetch_status_change(self):
+        """Called when fetch status changes."""
+
+
+class ItemTypeFetchParent(FetchParent):
+    def __init__(self, fetch_item_type):
+        self._fetch_item_type = fetch_item_type
+
+    @property
+    def fetch_item_type(self):
+        return self._fetch_item_type
+
+
+def load_project_dict(project_config_dir, logger):
+    """Loads project dictionary from project directory.
+
+    Args:
+        project_config_dir (str): project's .spinetoolbox directory
+        logger (LoggerInterface): a logger
+
+    Returns:
+        dict: project dictionary
+    """
+    load_path = os.path.abspath(os.path.join(project_config_dir, PROJECT_FILENAME))
+    try:
+        with open(load_path, "r") as fh:
+            try:
+                project_dict = json.load(fh)
+            except json.decoder.JSONDecodeError:
+                logger.msg_error.emit(f"Error in project file <b>{load_path}</b>. Invalid JSON.")
+                return None
+    except OSError:
+        logger.msg_error.emit(f"Project file <b>{load_path}</b> missing")
+        return None
+    return project_dict
+
+
+def load_local_project_data(project_config_dir, logger):
+    """Loads local project data.
+
+    Args:
+        project_config_dir (Path or str): project's .spinetoolbox directory
+        logger (LoggerInterface): a logger
+
+    Returns:
+        dict: project's local data
+    """
+    load_path = pathlib.Path(project_config_dir, PROJECT_LOCAL_DATA_DIR_NAME, PROJECT_LOCAL_DATA_FILENAME)
+    if not load_path.exists():
+        return {}
+    with load_path.open() as fh:
+        try:
+            local_data_dict = json.load(fh)
+        except json.decoder.JSONDecodeError:
+            logger.msg_error.emit(f"Error in project's local data file <b>{load_path}</b>. Invalid JSON.")
+            return {}
+    return local_data_dict
+
+
+def merge_dicts(source, target):
+    """Merges two dictionaries that may contain nested dictionaries recursively."""
+    for key, value in source.items():
+        target_entry = target.get(key)
+        if isinstance(value, dict) and target_entry is not None:
+            merge_dicts(value, target_entry)
+        else:
+            target[key] = value
+
+
+def fix_lightness_color(color, lightness=240):
+    h, s, _, a = color.getHsl()
+    return QColor.fromHsl(h, s, lightness, a)
+
+
+@contextmanager
+def scrolling_to_bottom(widget, tolerance=1):
+    scrollbar = widget.verticalScrollBar()
+    at_bottom = scrollbar.value() >= scrollbar.maximum() - tolerance
+    try:
+        yield None
+    finally:
+        if at_bottom:
+            scrollbar.setValue(scrollbar.maximum())
+
+
+def _is_metadata_item(item):
+    """Identifies a database metadata record.
+
+    Args:
+        item (dict): database item
+
+    Returns:
+        bool: True if item is metadata item, False otherwise
+    """
+    return "name" in item and "value" in item
+
+
+def separate_metadata_and_item_metadata(db_map_data):
+    """Separates normal metadata items from item metadata items.
+
+    Args:
+        db_map_data (dict): database records
+
+    Returns:
+        tuple: item metadata records and metadata records
+    """
+    metadata_db_map_data = {}
+    item_metadata_db_map_data = {}
+    for db_map, items in db_map_data.items():
+        metadata_items = []
+        entity_metadata_items = []
+        for item in items:
+            if _is_metadata_item(item):
+                metadata_items.append(item)
+            else:
+                entity_metadata_items.append(item)
+        if metadata_items:
+            metadata_db_map_data[db_map] = metadata_items
+        item_metadata_db_map_data[db_map] = entity_metadata_items
+    return item_metadata_db_map_data, metadata_db_map_data
