@@ -15,30 +15,29 @@ Models to represent items in a tree.
 :authors: P. Vennström (VTT), M. Marin (KTH)
 :date:   11.3.2019
 """
-from PySide2.QtCore import Qt, QAbstractItemModel, QModelIndex, QObject, Signal
+from PySide2.QtCore import Qt, QAbstractItemModel, QModelIndex
 
 
-class TreeItem(QObject):
+class TreeItem:
     """A tree item that can fetch its children."""
 
-    fully_fetched = Signal()
-
     def __init__(self, model=None):
-        """Initializes item.
-
+        """
         Args:
-            model (MinimalTreeModel, NoneType): The model where the item belongs.
+            model (MinimalTreeModel, optional): The model where the item belongs.
         """
         super().__init__()
+        self._children = []
         self._model = model
-        self._children = None
         self._parent_item = None
         self._fetched = False
-        self.children = []
-        self.fully_fetched.connect(self._handle_fully_fetched)
+        self._finalized = False
 
-    def _handle_fully_fetched(self):
-        """Handles fully_fetched."""
+    def has_children(self):
+        """Returns whether or not this item has or could have children."""
+        if self.can_fetch_more():
+            return True
+        return bool(self.child_count())
 
     @property
     def model(self):
@@ -71,12 +70,13 @@ class TreeItem(QObject):
         if not isinstance(parent_item, TreeItem) and parent_item is not None:
             raise ValueError("Parent must be instance of TreeItem or None")
         self._parent_item = parent_item
-        self._model = parent_item.model
+        if parent_item is not None:
+            self._model = parent_item.model
 
     def child(self, row):
         """Returns the child at given row or None if out of bounds."""
         try:
-            return self._children[row]
+            return self.children[row]
         except IndexError:
             return None
 
@@ -86,7 +86,7 @@ class TreeItem(QObject):
 
     def child_count(self):
         """Returns the number of children."""
-        return len(self._children)
+        return len(self.children)
 
     def child_number(self):
         """Returns the rank of this item within its parent or -1 if it's an orphan."""
@@ -117,6 +117,14 @@ class TreeItem(QObject):
     def index(self):
         return self.model.index_from_item(self)
 
+    def finalize(self):
+        if not self._finalized:
+            self._do_finalize()
+            self._finalized = True
+
+    def _do_finalize(self):
+        """Do some final initialization after setting the parent."""
+
     def insert_children(self, position, children):
         """Insert new children at given position. Returns a boolean depending on how it went.
 
@@ -132,8 +140,10 @@ class TreeItem(QObject):
         for child in children:
             child.parent_item = self
         self.model.beginInsertRows(self.index(), position, position + len(children) - 1)
-        self._children[position:position] = children
+        self.children[position:position] = children
         self.model.endInsertRows()
+        for child in children:
+            child.finalize()
         return True
 
     def append_children(self, children):
@@ -157,7 +167,9 @@ class TreeItem(QObject):
         if last >= self.child_count():
             last = self.child_count() - 1
         self.model.beginRemoveRows(self.index(), first, last)
-        del self._children[first : last + 1]
+        for child in self.children[first : last + 1]:
+            child.parent_item = None
+        del self.children[first : last + 1]
         self.model.endRemoveRows()
         return True
 
@@ -175,12 +187,6 @@ class TreeItem(QObject):
         """Returns data for given column and role."""
         return None
 
-    def has_children(self):
-        """Returns whether or not this item has or could have children."""
-        if self.child_count() or self.can_fetch_more():
-            return True
-        return False
-
     def can_fetch_more(self):
         """Returns whether or not this item can fetch more."""
         return not self._fetched
@@ -225,9 +231,18 @@ class MinimalTreeModel(QAbstractItemModel):
         self._parent = parent
         self._invisible_root_item = TreeItem(self)
 
-    def visit_all(self, index=QModelIndex()):
+    def visit_all(self, index=QModelIndex(), view=None):
         """Iterates all items in the model including and below the given index.
         Iterative implementation so we don't need to worry about Python recursion limits.
+
+        Args:
+            index (QModelIndex): an index to start. If not given, we start at the root
+            view (QTreeView): a tree view. If given, we only yield items that are visible
+                to that view. So for example, if a tree item is not expanded then we don't yield
+                its children.
+
+        Yields:
+            TreeItem
         """
         if index.isValid():
             ancient_one = self.item_from_index(index)
@@ -242,10 +257,11 @@ class MinimalTreeModel(QAbstractItemModel):
         while True:
             if visit_children:
                 yield current
-                child = current.last_child()
-                if child:
-                    current = child
-                    continue
+                if view is None or view.isExpanded(self.index_from_item(current)):
+                    child = current.last_child()
+                    if child:
+                        current = child
+                        continue
             sibling = current.previous_sibling()
             if sibling:
                 visit_children = True
@@ -314,8 +330,7 @@ class MinimalTreeModel(QAbstractItemModel):
         return True
 
     def flags(self, index):
-        """Returns the item flags for the given index.
-        """
+        """Returns the item flags for the given index."""
         item = self.item_from_index(index)
         return item.flags(index.column())
 
