@@ -16,26 +16,38 @@ Classes for drawing graphics items on QGraphicsScene.
 :date:    4.4.2018
 """
 
-from PySide2.QtCore import Qt, QPointF, QRectF, QParallelAnimationGroup
+import math
+from PySide2.QtCore import Qt, QPointF, QRectF, QLineF
 from PySide2.QtWidgets import (
     QGraphicsItem,
     QGraphicsTextItem,
     QGraphicsSimpleTextItem,
-    QGraphicsRectItem,
+    QGraphicsPathItem,
     QGraphicsEllipseItem,
     QGraphicsColorizeEffect,
     QGraphicsDropShadowEffect,
     QApplication,
     QToolTip,
+    QStyle,
+)
+from PySide2.QtGui import (
+    QColor,
+    QPen,
+    QBrush,
+    QTextCursor,
+    QPalette,
+    QTextBlockFormat,
+    QFont,
+    QPainterPath,
+    QRadialGradient,
 )
-from PySide2.QtGui import QColor, QPen, QBrush, QTextCursor, QPalette, QTextBlockFormat, QFont
 from PySide2.QtSvg import QGraphicsSvgItem, QSvgRenderer
 from spine_engine.spine_engine import ItemExecutionFinishState
 from .project_commands import MoveIconCommand
-from .helpers import LinkType
+from .helpers import LinkType, fix_lightness_color
 
 
-class ProjectItemIcon(QGraphicsRectItem):
+class ProjectItemIcon(QGraphicsPathItem):
     """Base class for project item icons drawn in Design View."""
 
     ITEM_EXTENT = 64
@@ -49,38 +61,71 @@ class ProjectItemIcon(QGraphicsRectItem):
         """
         super().__init__()
         self._toolbox = toolbox
+        self._scene = None
+        self._bumping = True
+        self.bumped_rects = {}  # Item rect before it was bumped
         self.icon_file = icon_file
+        self._icon_color = icon_color
         self._moved_on_scene = False
         self.previous_pos = QPointF()
-        self.current_pos = QPointF()
         self.icon_group = {self}
         self.renderer = QSvgRenderer()
         self.svg_item = QGraphicsSvgItem(self)
+        self.svg_item.setZValue(100)
         self.colorizer = QGraphicsColorizeEffect()
-        self.setRect(QRectF(-self.ITEM_EXTENT / 2, -self.ITEM_EXTENT / 2, self.ITEM_EXTENT, self.ITEM_EXTENT))
         self.text_font_size = 10  # point size
-        # Make item name graphics item.
-        self._name = ""
-        self.name_item = QGraphicsSimpleTextItem(self._name, self)
-        self.set_name_attributes()  # Set font, size, position, etc.
-        # Make connector buttons
-        self.connectors = dict(
-            bottom=ConnectorButton(self, toolbox, position="bottom"),
-            left=ConnectorButton(self, toolbox, position="left"),
-            right=ConnectorButton(self, toolbox, position="right"),
-        )
-        # Make exclamation and rank icons
+        self._rect = QRectF(-self.ITEM_EXTENT / 2, -self.ITEM_EXTENT / 2, self.ITEM_EXTENT, self.ITEM_EXTENT)
+        self.component_rect = QRectF(0, 0, self.ITEM_EXTENT / 4, self.ITEM_EXTENT / 4)
+        self._selection_halo = QGraphicsPathItem(self)
+        # Make exclamation, rank, and execution icons
         self.exclamation_icon = ExclamationIcon(self)
         self.execution_icon = ExecutionIcon(self)
         self.rank_icon = RankIcon(self)
-        h, s, _, a = icon_color.getHsl()
-        background_color = QColor.fromHsl(h, s, 240, a)
-        brush = QBrush(background_color)
-        self._setup(brush, icon_file, icon_color)
+        # Make item name graphics item.
+        self._name = ""
+        self.name_item = QGraphicsSimpleTextItem(self._name)
+        self.name_item.setZValue(100)
+        self.set_name_attributes()  # Set font, size, position, etc.
+        # Make connector buttons
+        self.connectors = dict(
+            bottom=ConnectorButton(toolbox, self, position="bottom"),
+            left=ConnectorButton(toolbox, self, position="left"),
+            right=ConnectorButton(toolbox, self, position="right"),
+        )
+        self._setup()
         shadow_effect = QGraphicsDropShadowEffect()
         shadow_effect.setOffset(1)
         shadow_effect.setEnabled(False)
         self.setGraphicsEffect(shadow_effect)
+        self._update_path()
+
+    def rect(self):
+        return self._rect
+
+    def _update_path(self):
+        rounded = self._toolbox.qsettings().value("appSettings/roundedItems", defaultValue="false") == "true"
+        self._do_update_path(rounded)
+
+    def update_path(self, rounded):
+        self._do_update_path(rounded)
+
+    def _do_update_path(self, rounded):
+        radius = self.component_rect.width() / 2 if rounded else 0
+        path = QPainterPath()
+        path.addRoundedRect(self._rect, radius, radius)
+        self.setPath(path)
+        self.rank_icon.update_path(radius)
+        for conn in self.connectors.values():
+            conn.update_path(radius)
+        # Selection halo
+        pen_width = 1
+        margin = 1
+        path = QPainterPath()
+        path.addRoundedRect(self._rect.adjusted(-margin, -margin, margin, margin), radius + margin, radius + margin)
+        self._selection_halo.setPath(path)
+        selection_pen = QPen(Qt.DashLine)
+        selection_pen.setWidthF(pen_width)
+        self._selection_halo.setPen(selection_pen)
 
     def finalize(self, name, x, y):
         """
@@ -91,24 +136,27 @@ class ProjectItemIcon(QGraphicsRectItem):
             x (int): horizontal offset
             y (int): vertical offset
         """
-        self.update_name_item(name)
         self.moveBy(x, y)
+        self.update_name_item(name)
 
-    def _setup(self, brush, svg, svg_color):
-        """Setup item's attributes.
-
-        Args:
-            brush (QBrush): Used in filling the background rectangle
-            svg (str): Path to SVG icon file
-            svg_color (QColor): Color of SVG icon
-        """
-        self.setPen(QPen(Qt.black, 1, Qt.SolidLine))
+    def _setup(self):
+        """Setup item's attributes."""
+        self.colorizer.setColor(self._icon_color)
+        background_color = fix_lightness_color(self._icon_color)
+        gradient = QRadialGradient(self._rect.center(), 1 * self._rect.width())
+        gradient.setColorAt(0, background_color.lighter(105))
+        gradient.setColorAt(1, background_color.darker(105))
+        brush = QBrush(gradient)
+        pen = QPen(QBrush(background_color.darker()), 1, Qt.SolidLine)
+        self.setPen(pen)
+        for conn in self.connectors.values():
+            conn.setPen(pen)
+        self.rank_icon.bg.setPen(pen)
         self.setBrush(brush)
-        self.colorizer.setColor(svg_color)
         # Load SVG
-        loading_ok = self.renderer.load(svg)
+        loading_ok = self.renderer.load(self.icon_file)
         if not loading_ok:
-            self._toolbox.msg_error.emit("Loading SVG icon from resource:{0} failed".format(svg))
+            self._toolbox.msg_error.emit("Loading SVG icon from resource:{0} failed".format(self.icon_file))
             return
         size = self.renderer.defaultSize()
         self.svg_item.setSharedRenderer(self.renderer)
@@ -148,7 +196,7 @@ class ProjectItemIcon(QGraphicsRectItem):
         """
         self._name = new_name
         self.name_item.setText(new_name)
-        self.set_name_attributes()
+        self._reposition_name_item()
 
     def set_name_attributes(self):
         """Set name QGraphicsSimpleTextItem attributes (font, size, position, etc.)"""
@@ -157,12 +205,12 @@ class ProjectItemIcon(QGraphicsRectItem):
         font.setPointSize(self.text_font_size)
         font.setBold(True)
         self.name_item.setFont(font)
-        # Set name item position (centered on top of the master icon)
-        name_width = self.name_item.boundingRect().width()
-        name_height = self.name_item.boundingRect().height()
-        self.name_item.setPos(
-            self.rect().x() + self.rect().width() / 2 - name_width / 2, self.rect().y() - name_height - 4
-        )
+
+    def _reposition_name_item(self):
+        """Set name item position (centered on top of the master icon)."""
+        main_rect = self.sceneBoundingRect()
+        name_rect = self.name_item.sceneBoundingRect()
+        self.name_item.setPos(main_rect.center().x() - name_rect.width() / 2, main_rect.y() - name_rect.height() - 4)
 
     def conn_button(self, position="left"):
         """Returns item's connector button.
@@ -191,19 +239,6 @@ class ProjectItemIcon(QGraphicsRectItem):
         """
         return [l for conn in self.connectors.values() for l in conn.incoming_links()]
 
-    def run_execution_leave_animation(self, excluded):
-        """
-        Starts the animation associated with execution leaving the icon.
-
-        Args:
-            excluded (bool): True if project item was not actually executed.
-        """
-        animation_group = QParallelAnimationGroup()
-        for link in self.outgoing_connection_links():
-            animation_group.addAnimation(link.make_execution_animation(excluded))
-        animation_group.finished.connect(animation_group.deleteLater)
-        animation_group.start()
-
     def hoverEnterEvent(self, event):
         """Sets a drop shadow effect to icon when mouse enters its boundaries.
 
@@ -225,36 +260,29 @@ class ProjectItemIcon(QGraphicsRectItem):
         event.accept()
 
     def mousePressEvent(self, event):
+        """Updates scene's icon group."""
         super().mousePressEvent(event)
-        self.icon_group = set(x for x in self.scene().selectedItems() if isinstance(x, ProjectItemIcon)) | {self}
-        for icon in self.icon_group:
+        icon_group = set(x for x in self.scene().selectedItems() if isinstance(x, ProjectItemIcon)) | {self}
+        for icon in icon_group:
             icon.previous_pos = icon.scenePos()
-
-    def mouseMoveEvent(self, event):
-        """Moves icon(s) while the mouse button is pressed.
-        Update links that are connected to selected icons.
-
-        Args:
-            event (QGraphicsSceneMouseEvent): Event
-        """
-        super().mouseMoveEvent(event)
-        self.update_links_geometry()
-
-    def moveBy(self, dx, dy):
-        super().moveBy(dx, dy)
-        self.update_links_geometry()
+        self.scene().icon_group = icon_group
 
     def update_links_geometry(self):
         """Updates geometry of connected links to reflect this item's most recent position."""
-        links = set(link for icon in self.icon_group for conn in icon.connectors.values() for link in conn.links)
-        for link in links:
-            link.update_geometry()
+        scene = self.scene()
+        if not scene:
+            return
+        icon_group = scene.icon_group | {self}
+        scene.dirty_links |= set(
+            link for icon in icon_group for conn in icon.connectors.values() for link in conn.links
+        )
 
     def mouseReleaseEvent(self, event):
-        for icon in self.icon_group:
-            icon.current_pos = icon.scenePos()
+        """Clears pre-bump rects, and pushes a move icon command if necessary."""
+        for icon in self.scene().icon_group:
+            icon.bumped_rects.clear()
         # pylint: disable=undefined-variable
-        if (self.current_pos - self.previous_pos).manhattanLength() > qApp.startDragDistance():
+        if (self.scenePos() - self.previous_pos).manhattanLength() > qApp.startDragDistance():
             self._toolbox.undo_stack.push(MoveIconCommand(self, self._toolbox.project()))
             event.ignore()
         super().mouseReleaseEvent(event)
@@ -293,29 +321,98 @@ class ProjectItemIcon(QGraphicsRectItem):
         """
         if change == QGraphicsItem.ItemScenePositionHasChanged:
             self._moved_on_scene = True
+            self._reposition_name_item()
+            self.update_links_geometry()
+            self._handle_collisions()
         elif change == QGraphicsItem.GraphicsItemChange.ItemSceneChange and value is None:
             self.prepareGeometryChange()
             self.setGraphicsEffect(None)
+        elif change == QGraphicsItem.GraphicsItemChange.ItemSceneHasChanged:
+            scene = value
+            if scene is None:
+                self._scene.removeItem(self.name_item)
+            else:
+                self._scene = scene
+                self._scene.addItem(self.name_item)
+                self._reposition_name_item()
         return super().itemChange(change, value)
 
+    def set_pos_without_bumping(self, pos):
+        """Sets position without bumping other items. Needed for undoing move operations.
+
+        Args:
+            pos (QPointF)
+        """
+        self._bumping = False
+        self.setPos(pos)
+        self._bumping = True
+
+    def _handle_collisions(self):
+        """Handles collisions with other items."""
+        prevent_overlapping = self._toolbox.qsettings().value("appSettings/preventOverlapping", defaultValue="false")
+        if not self.scene() or not self._bumping or prevent_overlapping != "true":
+            return
+        restablished = self._restablish_bumped_items()
+        for other in set(self.collidingItems()) - restablished:
+            if isinstance(other, ProjectItemIcon):
+                other.make_room_for_item(self)
+
+    def make_room_for_item(self, other):
+        """Makes room for another item.
+
+        Args:
+            item (ProjectItemIcon)
+        """
+        if self not in other.bumped_rects:
+            other.bumped_rects[self] = self.sceneBoundingRect()
+            if self not in self.scene().icon_group:
+                self.scene().icon_group.add(self)
+                self.previous_pos = self.scenePos()
+        line = QLineF(other.sceneBoundingRect().center(), self.sceneBoundingRect().center())
+        intersection = other.sceneBoundingRect() & self.sceneBoundingRect()
+        delta = math.atan(line.angle()) * min(intersection.width(), intersection.height())
+        unit_vector = line.unitVector()
+        self.moveBy(delta * unit_vector.dx(), delta * unit_vector.dy())
+
+    def _restablish_bumped_items(self):
+        """Moves bumped items back to their original position if no collision would happen anymore."""
+        restablished = set()
+        try:
+            for other, rect in self.bumped_rects.items():
+                if not self.sceneBoundingRect().intersects(rect):
+                    other.setPos(rect.center())
+                    restablished.add(other)
+            for other in restablished:
+                self.bumped_rects.pop(other, None)
+        except RuntimeError:
+            pass
+        return restablished
+
     def select_item(self):
         """Update GUI to show the details of the selected item."""
         ind = self._toolbox.project_item_model.find_item(self.name())
         self._toolbox.ui.treeView_project.setCurrentIndex(ind)
 
+    def paint(self, painter, option, widget=None):
+        """Sets a dashed pen if selected."""
+        selected = option.state & QStyle.State_Selected
+        self._selection_halo.setVisible(selected)
+        option.state &= ~QStyle.State_Selected
+        super().paint(painter, option, widget)
 
-class ConnectorButton(QGraphicsRectItem):
+
+class ConnectorButton(QGraphicsPathItem):
     """Connector button graphics item. Used for Link drawing between project items."""
 
     # Regular and hover brushes
     brush = QBrush(QColor(255, 255, 255))  # Used in filling the item
     hover_brush = QBrush(QColor(50, 0, 50, 128))  # Used in filling the item while hovering
 
-    def __init__(self, parent, toolbox, position="left"):
+    def __init__(self, toolbox, parent, position="left"):
         """
         Args:
-            parent (QGraphicsItem): Project item bg rectangle
             toolbox (ToolboxUI): QMainWindow instance
+            parent (ProjectItemIcon): parent graphics item
             position (str): Either "top", "left", "bottom", or "right"
         """
         super().__init__(parent)
@@ -323,24 +420,30 @@ class ConnectorButton(QGraphicsRectItem):
         self._toolbox = toolbox
         self.position = position
         self.links = list()
-        pen = QPen(Qt.black, 0.5, Qt.SolidLine)
-        self.setPen(pen)
         self.setBrush(self.brush)
         parent_rect = parent.rect()
         extent = 0.2 * parent_rect.width()
-        rect = QRectF(0, 0, extent, extent)
+        self._rect = QRectF(0, 0, extent, extent)
         if position == "top":
-            rect.moveCenter(QPointF(parent_rect.center().x(), parent_rect.top() + extent / 2))
+            self._rect.moveCenter(QPointF(parent_rect.center().x(), parent_rect.top() + extent / 2))
         elif position == "left":
-            rect.moveCenter(QPointF(parent_rect.left() + extent / 2, parent_rect.center().y()))
+            self._rect.moveCenter(QPointF(parent_rect.left() + extent / 2, parent_rect.center().y()))
         elif position == "bottom":
-            rect.moveCenter(QPointF(parent_rect.center().x(), parent_rect.bottom() - extent / 2))
+            self._rect.moveCenter(QPointF(parent_rect.center().x(), parent_rect.bottom() - extent / 2))
         elif position == "right":
-            rect.moveCenter(QPointF(parent_rect.right() - extent / 2, parent_rect.center().y()))
-        self.setRect(rect)
+            self._rect.moveCenter(QPointF(parent_rect.right() - extent / 2, parent_rect.center().y()))
         self.setAcceptHoverEvents(True)
         self.setCursor(Qt.PointingHandCursor)
 
+    def rect(self):
+        return self._rect
+
+    def update_path(self, parent_radius):
+        radius = 0.2 * parent_radius
+        path = QPainterPath()
+        path.addRoundedRect(self._rect, radius, radius)
+        self.setPath(path)
+
     @property
     def parent(self):
         return self._parent
@@ -572,9 +675,8 @@ class RankIcon(QGraphicsTextItem):
         """
         super().__init__(parent)
         self._parent = parent
-        rect_w = parent.rect().width()  # Parent rect width
-        self.text_margin = 0.05 * rect_w
-        self.bg = QGraphicsRectItem(self.boundingRect(), self)
+        self._rect = parent.component_rect
+        self.bg = QGraphicsPathItem(self)
         bg_brush = QApplication.palette().brush(QPalette.ToolTipBase)
         self.bg.setBrush(bg_brush)
         self.bg.setFlag(QGraphicsItem.ItemStacksBehindParent)
@@ -586,11 +688,27 @@ class RankIcon(QGraphicsTextItem):
         doc = self.document()
         doc.setDocumentMargin(0)
 
+    def _make_path(self, radius):
+        path = QPainterPath()
+        if radius == 0:
+            path.addRect(self._rect)
+            return path
+        path.moveTo(0, self._rect.height())
+        path.lineTo(0.5 * self._rect.width(), self._rect.height())
+        path.arcTo(self._rect, 270, 90)
+        path.lineTo(self._rect.width(), 0)
+        path.lineTo(0.5 * self._rect.width(), 0)
+        path.arcTo(self._rect, 90, 90)
+        path.lineTo(0, self._rect.height())
+        return path
+
+    def update_path(self, radius):
+        path = self._make_path(radius)
+        self.bg.setPath(path)
+
     def set_rank(self, rank):
         self.setPlainText(str(rank))
-        self.adjustSize()
-        self.setTextWidth(self.text_margin + self.textWidth())
-        self.bg.setRect(self.boundingRect())
+        self.setTextWidth(self._rect.width())
         # Align center
         fmt = QTextBlockFormat()
         fmt.setAlignment(Qt.AlignHCenter)
