@@ -17,8 +17,8 @@ Custom item delegates.
 """
 
 from numbers import Number
-from PySide2.QtCore import QModelIndex, QPoint, Qt, Signal
-from PySide2.QtWidgets import QStyledItemDelegate
+from PySide2.QtCore import QModelIndex, QPoint, Qt, Signal, QSortFilterProxyModel, QRegExp
+from PySide2.QtWidgets import QStyledItemDelegate, QComboBox
 from PySide2.QtGui import QFontMetrics
 from spinedb_api import to_database
 from spinedb_api.parameter_value import join_value_and_type
@@ -26,6 +26,9 @@ from ...widgets.custom_editors import CustomLineEditor, SearchBarEditor, CheckLi
 from ...mvcmodels.shared import PARSED_ROLE
 from ...widgets.custom_delegates import CheckBoxDelegate, RankDelegate
 from ...helpers import object_icon
+from ..mvcmodels.metadata_table_model_base import Column as MetadataColumn
+
+# FIXME: only_visible=False ???
 
 
 class RelationshipPivotTableDelegate(CheckBoxDelegate):
@@ -236,11 +239,7 @@ class ParameterDelegate(QStyledItemDelegate):
     def updateEditorGeometry(self, editor, option, index):
         super().updateEditorGeometry(editor, option, index)
         if isinstance(editor, (SearchBarEditor, CheckListEditor)):
-            size = option.rect.size()
-            if index.data(Qt.DecorationRole):
-                size.setWidth(size.width() - 22)  # FIXME
-            editor.set_base_size(size)
-            editor.update_geometry()
+            editor.update_geometry(option)
 
     def _close_editor(self, editor, index):
         """Closes editor. Needed by SearchBarEditor."""
@@ -270,7 +269,7 @@ class DatabaseNameDelegate(ParameterDelegate):
 
 
 class ParameterValueOrDefaultValueDelegate(ParameterDelegate):
-    """A delegate for the either the value or the default value."""
+    """A delegate for either the value or the default value."""
 
     parameter_value_editor_requested = Signal(QModelIndex)
 
@@ -300,13 +299,18 @@ class ParameterValueOrDefaultValueDelegate(ParameterDelegate):
         """If the parameter has associated a value list, returns a SearchBarEditor.
         Otherwise returns or requests a dedicated parameter_value editor.
         """
+        self._db_value_list_lookup = {}
         db_map = self._get_db_map(index)
         if not db_map:
             return None
         value_list_id = self._get_value_list_id(index, db_map)
         if value_list_id:
-            display_value_list = self.db_mngr.get_parameter_value_list(db_map, value_list_id, Qt.DisplayRole)
-            db_value_list = self.db_mngr.get_parameter_value_list(db_map, value_list_id, Qt.EditRole)
+            display_value_list = self.db_mngr.get_parameter_value_list(
+                db_map, value_list_id, Qt.DisplayRole, only_visible=False
+            )
+            db_value_list = self.db_mngr.get_parameter_value_list(
+                db_map, value_list_id, Qt.EditRole, only_visible=False
+            )
             self._db_value_list_lookup = dict(zip(display_value_list, db_value_list))
             editor = SearchBarEditor(self.parent(), parent)
             editor.set_data(index.data(), self._db_value_list_lookup)
@@ -322,7 +326,9 @@ class ParameterDefaultValueDelegate(ParameterValueOrDefaultValueDelegate):
         """Returns a value list item for the given index and db_map."""
         h = index.model().header.index
         value_list_name = index.sibling(index.row(), h("value_list_name")).data()
-        value_lists = self.db_mngr.get_items_by_field(db_map, "parameter_value_list", "name", value_list_name)
+        value_lists = self.db_mngr.get_items_by_field(
+            db_map, "parameter_value_list", "name", value_list_name, only_visible=False
+        )
         if len(value_lists) == 1:
             return value_lists[0]["id"]
 
@@ -334,21 +340,21 @@ class ParameterValueDelegate(ParameterValueOrDefaultValueDelegate):
         """Returns a value list item for the given index and db_map."""
         h = index.model().header.index
         parameter_name = index.sibling(index.row(), h("parameter_name")).data()
-        parameters = self.db_mngr.get_items_by_field(db_map, "parameter_definition", "parameter_name", parameter_name)
+        parameters = self.db_mngr.get_items_by_field(
+            db_map, "parameter_definition", "parameter_name", parameter_name, only_visible=False
+        )
         entity_class_id = index.model().get_entity_class_id(index, db_map)
         parameter_ids = {p["id"] for p in parameters if p["entity_class_id"] == entity_class_id}
         value_list_ids = {
-            self.db_mngr.get_item(db_map, "parameter_definition", id_).get("value_list_id") for id_ in parameter_ids
+            self.db_mngr.get_item(db_map, "parameter_definition", id_, only_visible=False).get("value_list_id")
+            for id_ in parameter_ids
         }
         if len(value_list_ids) == 1:
             return next(iter(value_list_ids))
 
 
-# FIXME: Make SearchBarEditor lazy
-
-
 class ValueListDelegate(ParameterDelegate):
-    """A delegate for the parameter_value-list."""
+    """A delegate for the parameter value list."""
 
     def createEditor(self, parent, option, index):
         """Returns editor."""
@@ -356,7 +362,7 @@ class ValueListDelegate(ParameterDelegate):
         if not db_map:
             return None
         editor = SearchBarEditor(self.parent(), parent)
-        name_list = [x["name"] for x in self.db_mngr.get_items(db_map, "parameter_value_list")]
+        name_list = [x["name"] for x in self.db_mngr.get_items(db_map, "parameter_value_list", only_visible=False)]
         editor.set_data(index.data(Qt.EditRole), name_list)
         editor.data_committed.connect(lambda editor=editor, index=index: self._close_editor(editor, index))
         return editor
@@ -371,7 +377,7 @@ class ObjectClassNameDelegate(ParameterDelegate):
         if not db_map:
             return None
         editor = SearchBarEditor(self.parent(), parent)
-        object_classes = self.db_mngr.get_items(db_map, "object_class")
+        object_classes = self.db_mngr.get_items(db_map, "object_class", only_visible=False)
         editor.set_data(index.data(Qt.EditRole), [x["name"] for x in object_classes])
         editor.data_committed.connect(lambda editor=editor, index=index: self._close_editor(editor, index))
         return editor
@@ -386,7 +392,7 @@ class RelationshipClassNameDelegate(ParameterDelegate):
         if not db_map:
             return None
         editor = SearchBarEditor(self.parent(), parent)
-        relationship_classes = self.db_mngr.get_items(db_map, "relationship_class")
+        relationship_classes = self.db_mngr.get_items(db_map, "relationship_class", only_visible=False)
         editor.set_data(index.data(Qt.EditRole), [x["name"] for x in relationship_classes])
         editor.data_committed.connect(lambda editor=editor, index=index: self._close_editor(editor, index))
         return editor
@@ -404,10 +410,10 @@ class ParameterNameDelegate(ParameterDelegate):
         entity_class_id = index.model().get_entity_class_id(index, db_map)
         if entity_class_id is not None:
             parameter_definitions = self.db_mngr.get_items_by_field(
-                db_map, "parameter_definition", "entity_class_id", entity_class_id
+                db_map, "parameter_definition", "entity_class_id", entity_class_id, only_visible=False
             )
         else:
-            parameter_definitions = self.db_mngr.get_items(db_map, "parameter_definition")
+            parameter_definitions = self.db_mngr.get_items(db_map, "parameter_definition", only_visible=False)
         name_list = list({x["parameter_name"]: None for x in parameter_definitions})
         editor.set_data(index.data(Qt.EditRole), name_list)
         editor.data_committed.connect(lambda editor=editor, index=index: self._close_editor(editor, index))
@@ -425,9 +431,9 @@ class ObjectNameDelegate(ParameterDelegate):
         editor = SearchBarEditor(self.parent(), parent)
         object_class_id = index.model().get_entity_class_id(index, db_map)
         if object_class_id is not None:
-            objects = self.db_mngr.get_items_by_field(db_map, "object", "class_id", object_class_id)
+            objects = self.db_mngr.get_items_by_field(db_map, "object", "class_id", object_class_id, only_visible=False)
         else:
-            objects = self.db_mngr.get_items(db_map, "object")
+            objects = self.db_mngr.get_items(db_map, "object", only_visible=False)
         name_list = list({x["name"]: None for x in objects})
         editor.set_data(index.data(Qt.EditRole), name_list)
         editor.data_committed.connect(lambda editor=editor, index=index: self._close_editor(editor, index))
@@ -443,7 +449,7 @@ class AlternativeNameDelegate(ParameterDelegate):
         if not db_map:
             return None
         editor = SearchBarEditor(self.parent(), parent)
-        name_list = [x["name"] for x in self.db_mngr.get_items(db_map, "alternative")]
+        name_list = [x["name"] for x in self.db_mngr.get_items(db_map, "alternative", only_visible=False)]
         editor.set_data(index.data(Qt.EditRole), name_list)
         editor.data_committed.connect(lambda editor=editor, index=index: self._close_editor(editor, index))
         return editor
@@ -452,7 +458,7 @@ class AlternativeNameDelegate(ParameterDelegate):
 class ObjectNameListDelegate(ParameterDelegate):
     """A delegate for the object name list."""
 
-    object_name_list_editor_requested = Signal("QModelIndex", int, "QVariant")
+    object_name_list_editor_requested = Signal(QModelIndex, int, object)
 
     def createEditor(self, parent, option, index):
         """Returns editor."""
@@ -472,14 +478,27 @@ class ToolFeatureDelegate(QStyledItemDelegate):
 
     data_committed = Signal(QModelIndex, object)
 
-    @staticmethod
-    def _get_names(item, model):
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        self._feature_ids = {}
+
+    def _get_names(self, item, model):
         if item.item_type == "feature":
             names = model.get_all_feature_names(item.db_map)
             if not names:
-                model._parent.msg_error.emit("There aren't any listed parameter definitions to create features from. ")
+                model._parent.msg_error.emit("There aren't any listed parameter definitions to create features from.")
                 return None
             return names
+        if item.item_type == "tool_feature":
+            self._feature_ids = {
+                model.make_feature_name(x["entity_class_name"], x["parameter_definition_name"]): (
+                    x["id"],
+                    x["parameter_value_list_id"],
+                )
+                for x in item.db_mngr.get_items(item.db_map, "feature", only_visible=False)
+                if x["id"] not in item.parent_item.feature_id_list
+            }
+            return list(self._feature_ids)
         if item.item_type == "tool_feature required":
             return ["yes", "no"]
         if item.item_type == "tool_feature_method":
@@ -499,9 +518,14 @@ class ToolFeatureDelegate(QStyledItemDelegate):
         """Send signal."""
         item = index.model().item_from_index(index)
         index_data = self._get_index_data(item, index)
-        if editor.data() == index_data:
+        editor_data = editor.data()
+        if editor_data == index_data:
             return
-        self.data_committed.emit(index, editor.data())
+        if item.item_type == "tool_feature":
+            editor_data = self._feature_ids.get(editor_data)
+            if editor_data is None:
+                return
+        self.data_committed.emit(index, editor_data)
 
     def setEditorData(self, editor, index):
         """Do nothing. We're setting editor data right away in createEditor."""
@@ -510,7 +534,7 @@ class ToolFeatureDelegate(QStyledItemDelegate):
         """Returns editor."""
         model = index.model()
         item = model.item_from_index(index)
-        if item.item_type in ("feature", "tool_feature required", "tool_feature_method"):
+        if item.item_type in ("feature", "tool_feature", "tool_feature required", "tool_feature_method"):
             editor = SearchBarEditor(self.parent(), parent)
             index_data = self._get_index_data(item, index)
             names = self._get_names(item, model)
@@ -518,22 +542,19 @@ class ToolFeatureDelegate(QStyledItemDelegate):
                 return None
             editor.set_data(index_data, names)
             editor.data_committed.connect(lambda editor=editor, index=index: self._close_editor(editor, index))
-            return editor
-        editor = CustomLineEditor(parent)
-        editor.set_data(index.data(Qt.EditRole))
+        else:
+            editor = CustomLineEditor(parent)
+            editor.set_data(index.data(Qt.EditRole))
         return editor
 
     def updateEditorGeometry(self, editor, option, index):
         super().updateEditorGeometry(editor, option, index)
         item = index.model().item_from_index(index)
-        if item.item_type in ("feature", "tool_feature required", "tool_feature_method"):
-            size = option.rect.size()
+        if item.item_type in ("feature", "tool_feature", "tool_feature required", "tool_feature_method"):
             if item.item_type == "tool_feature required":
                 dx = QFontMetrics(index.data(Qt.FontRole)).horizontalAdvance("required:")
-                size.setWidth(size.width() - dx)
                 editor.set_base_offset(QPoint(dx, 0))
-            editor.set_base_size(size)
-            editor.update_geometry()
+            editor.update_geometry(option)
 
     def _close_editor(self, editor, index):
         """Closes editor. Needed by SearchBarEditor."""
@@ -546,40 +567,69 @@ class AlternativeScenarioDelegate(QStyledItemDelegate):
 
     data_committed = Signal(QModelIndex, object)
 
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        self._alternative_ids = {}
+
     def setModelData(self, editor, model, index):
         """Send signal."""
-        if editor.data() == index.data():
+        item = index.model().item_from_index(index)
+        index_data = self._get_index_data(item, index)
+        editor_data = editor.data()
+        if editor_data == index_data:
             return
-        self.data_committed.emit(index, editor.data())
+        if item.item_type == "scenario_alternative":
+            editor_data = self._alternative_ids.get(editor_data)
+            if editor_data is None:
+                return
+        self.data_committed.emit(index, editor_data)
 
     def setEditorData(self, editor, index):
         """Do nothing. We're setting editor data right away in createEditor."""
 
+    def _get_names(self, item, model):
+        if item.item_type == "scenario_alternative":
+            self._alternative_ids = {
+                x["name"]: x["id"]
+                for x in item.db_mngr.get_items(item.db_map, "alternative", only_visible=False)
+                if x["id"] not in item.parent_item.alternative_id_list
+            }
+            return list(self._alternative_ids)
+        if item.item_type == "scenario active":
+            return ["yes", "no"]
+
+    @staticmethod
+    def _get_index_data(item, index):
+        index_data = index.data(Qt.EditRole)
+        if item.item_type == "scenario active":
+            return index_data.split(": ")[1]
+        return index_data
+
     def createEditor(self, parent, option, index):
         """Returns editor."""
         model = index.model()
         item = model.item_from_index(index)
-        if item.item_type != "scenario active":
+        if item.item_type in ("scenario active", "scenario_alternative"):
+            editor = SearchBarEditor(self.parent(), parent)
+            index_data = self._get_index_data(item, index)
+            names = self._get_names(item, model)
+            if names is None:
+                return None
+            editor.set_data(index_data, names)
+            editor.data_committed.connect(lambda editor=editor, index=index: self._close_editor(editor, index))
+        else:
             editor = CustomLineEditor(parent)
             editor.set_data(index.data(Qt.EditRole))
-            return editor
-        editor = SearchBarEditor(self.parent(), parent)
-        active = index.data(Qt.EditRole).split(": ")[1]
-        editor.set_data(active, ["yes", "no"])
-        editor.data_committed.connect(lambda editor=editor, index=index: self._close_editor(editor, index))
         return editor
 
     def updateEditorGeometry(self, editor, option, index):
         super().updateEditorGeometry(editor, option, index)
         item = index.model().item_from_index(index)
-        if item.item_type == "scenario active":
-            size = option.rect.size()
-            fm = QFontMetrics(index.data(Qt.FontRole))
-            dx = fm.horizontalAdvance("active:")
-            size.setWidth(size.width() - dx)
-            editor.set_base_size(size)
-            editor.set_base_offset(QPoint(dx, 0))
-            editor.update_geometry()
+        if item.item_type in ("scenario active", "scenario_alternative"):
+            if item.item_type == "scenario active":
+                dx = QFontMetrics(index.data(Qt.FontRole)).horizontalAdvance("active:")
+                editor.set_base_offset(QPoint(dx, 0))
+            editor.update_geometry(option)
 
     def _close_editor(self, editor, index):
         """Closes editor. Needed by SearchBarEditor."""
@@ -595,9 +645,12 @@ class ParameterValueListDelegate(QStyledItemDelegate):
 
     def setModelData(self, editor, model, index):
         """Send signal."""
-        if editor.data() == index.data(Qt.EditRole):
-            return
-        self.data_committed.emit(index, editor.data())
+        self.closeEditor.emit(editor)
+        item = model.item_from_index(index)
+        data = editor.data()
+        if item.item_type == "list_value":
+            data = to_database(data)
+        self.data_committed.emit(index, data)
 
     def setEditorData(self, editor, index):
         """Do nothing. We're setting editor data right away in createEditor."""
@@ -606,7 +659,7 @@ class ParameterValueListDelegate(QStyledItemDelegate):
         """Returns editor."""
         model = index.model()
         item = model.item_from_index(index)
-        if item.item_type != "value":
+        if item.item_type != "list_value":
             editor = CustomLineEditor(parent)
             editor.set_data(index.data(Qt.EditRole))
             return editor
@@ -624,8 +677,7 @@ class ParameterValueListDelegate(QStyledItemDelegate):
 
 
 class ManageItemsDelegate(QStyledItemDelegate):
-    """A custom delegate for the model in {Add/Edit}ItemDialogs.
-    """
+    """A custom delegate for the model in {Add/Edit}ItemDialogs."""
 
     data_committed = Signal(QModelIndex, object)
 
@@ -640,15 +692,10 @@ class ManageItemsDelegate(QStyledItemDelegate):
     def updateEditorGeometry(self, editor, option, index):
         super().updateEditorGeometry(editor, option, index)
         if isinstance(editor, (SearchBarEditor, CheckListEditor)):
-            size = option.rect.size()
-            if index.data(Qt.DecorationRole):
-                size.setWidth(size.width() - 22)  # FIXME
-            editor.set_base_size(size)
-            editor.update_geometry()
+            editor.update_geometry(option)
 
     def connect_editor_signals(self, editor, index):
-        """Connect editor signals if necessary.
-        """
+        """Connect editor signals if necessary."""
         if isinstance(editor, SearchBarEditor):
             model = index.model()
             editor.data_committed.connect(lambda e=editor, i=index, m=model: self.close_editor(e, i, m))
@@ -672,9 +719,19 @@ class ManageItemsDelegate(QStyledItemDelegate):
         return editor
 
 
-class ManageObjectClassesDelegate(ManageItemsDelegate):
-    """A delegate for the model and view in {Add/Edit}ObjectClassesDialog.
-    """
+class ManageEntityClassesDelegate(ManageItemsDelegate):
+    def paint(self, painter, option, index):
+        """Get a pixmap from the index data and paint it in the middle of the cell."""
+        header = index.model().horizontal_header_labels()
+        if header[index.column()] == 'display icon':
+            icon = object_icon(index.data())
+            icon.paint(painter, option.rect, Qt.AlignVCenter | Qt.AlignHCenter)
+        else:
+            super().paint(painter, option, index)
+
+
+class ManageObjectClassesDelegate(ManageEntityClassesDelegate):
+    """A delegate for the model and view in {Add/Edit}ObjectClassesDialog."""
 
     icon_color_editor_requested = Signal(QModelIndex)
 
@@ -692,19 +749,9 @@ class ManageObjectClassesDelegate(ManageItemsDelegate):
         self.connect_editor_signals(editor, index)
         return editor
 
-    def paint(self, painter, option, index):
-        """Get a pixmap from the index data and paint it in the middle of the cell."""
-        header = index.model().horizontal_header_labels()
-        if header[index.column()] == 'display icon':
-            icon = object_icon(index.data())
-            icon.paint(painter, option.rect, Qt.AlignVCenter | Qt.AlignHCenter)
-        else:
-            super().paint(painter, option, index)
-
 
 class ManageObjectsDelegate(ManageItemsDelegate):
-    """A delegate for the model and view in {Add/Edit}ObjectsDialog.
-    """
+    """A delegate for the model and view in {Add/Edit}ObjectsDialog."""
 
     def createEditor(self, parent, option, index):
         """Return editor."""
@@ -722,14 +769,18 @@ class ManageObjectsDelegate(ManageItemsDelegate):
         return editor
 
 
-class ManageRelationshipClassesDelegate(ManageItemsDelegate):
-    """A delegate for the model and view in {Add/Edit}RelationshipClassesDialog.
-    """
+class ManageRelationshipClassesDelegate(ManageEntityClassesDelegate):
+    """A delegate for the model and view in {Add/Edit}RelationshipClassesDialog."""
+
+    icon_color_editor_requested = Signal(QModelIndex)
 
     def createEditor(self, parent, option, index):
         """Return editor."""
         header = index.model().horizontal_header_labels()
-        if header[index.column()] in ('relationship_class name', 'description'):
+        if header[index.column()] == 'display icon':
+            self.icon_color_editor_requested.emit(index)
+            editor = None
+        elif header[index.column()] in ('relationship_class name', 'description'):
             editor = CustomLineEditor(parent)
             editor.set_data(index.data(Qt.EditRole))
         elif header[index.column()] == 'databases':
@@ -743,8 +794,7 @@ class ManageRelationshipClassesDelegate(ManageItemsDelegate):
 
 
 class ManageRelationshipsDelegate(ManageItemsDelegate):
-    """A delegate for the model and view in {Add/Edit}RelationshipsDialog.
-    """
+    """A delegate for the model and view in {Add/Edit}RelationshipsDialog."""
 
     def createEditor(self, parent, option, index):
         """Return editor."""
@@ -764,8 +814,7 @@ class ManageRelationshipsDelegate(ManageItemsDelegate):
 
 
 class RemoveEntitiesDelegate(ManageItemsDelegate):
-    """A delegate for the model and view in RemoveEntitiesDialog.
-    """
+    """A delegate for the model and view in RemoveEntitiesDialog."""
 
     def createEditor(self, parent, option, index):
         """Return editor."""
@@ -774,3 +823,35 @@ class RemoveEntitiesDelegate(ManageItemsDelegate):
             editor = self._create_database_editor(parent, option, index)
             self.connect_editor_signals(editor, index)
             return editor
+
+
+class ItemMetadataDelegate(QStyledItemDelegate):
+    """A delegate for name and value columns in item metadata editor."""
+
+    def __init__(self, item_metadata_model, metadata_model, column, parent):
+        """
+        Args:
+            item_metadata_model (ItemMetadataModel): item metadata model
+            metadata_model (MetadataTableModel): metadata model
+            column (int): item metadata table column column
+            parent (QObject, optional): parent object
+        """
+        super().__init__(parent)
+        self._item_metadata_model = item_metadata_model
+        self._metadata_model = metadata_model
+        self._column = column
+
+    def createEditor(self, parent, option, index):
+        editor = QComboBox(parent)
+        editor.setEditable(True)
+        database_codename = self._item_metadata_model.index(index.row(), MetadataColumn.DB_MAP).data()
+        items = set()
+        if database_codename:
+            for i in range(self._metadata_model.rowCount() - 1):
+                if self._metadata_model.index(i, MetadataColumn.DB_MAP).data() == database_codename:
+                    items.add(self._metadata_model.index(i, self._column).data())
+        else:
+            for i in range(self._metadata_model.rowCount() - 1):
+                items.add(self._metadata_model.index(i, self._column).data())
+        editor.addItems(sorted(items))
+        return editor
