@@ -22,6 +22,8 @@ from PySide2.QtGui import QMouseEvent, QIcon
 from spinetoolbox.widgets.custom_qtreeview import CopyTreeView
 from spinetoolbox.helpers import busy_effect, CharIconEngine
 from .custom_delegates import ToolFeatureDelegate, AlternativeScenarioDelegate, ParameterValueListDelegate
+from .scenario_generator import ScenarioGenerator
+from ..mvcmodels.alternative_scenario_item import AlternativeLeafItem
 
 
 class EntityTreeView(CopyTreeView):
@@ -41,7 +43,6 @@ class EntityTreeView(CopyTreeView):
         self._add_relationship_classes_action = None
         self._add_relationships_action = None
         self._manage_relationships_action = None
-        self._show_entity_metadata_action = None
         self._export_action = None
         self._edit_action = None
         self._remove_action = None
@@ -55,6 +56,10 @@ class EntityTreeView(CopyTreeView):
         self._fetch_more_timer.setInterval(100)
         self._fetch_more_timer.timeout.connect(self._fetch_more_visible)
 
+    def reset(self):
+        super().reset()
+        self._selected_indexes = {}
+
     def connect_spine_db_editor(self, spine_db_editor):
         """Connects a Spine db editor to work with this view.
 
@@ -76,10 +81,6 @@ class EntityTreeView(CopyTreeView):
         self._menu.addSeparator()
         self._add_middle_actions()
         self._menu.addSeparator()
-        self._show_entity_metadata_action = self._menu.addAction(
-            QIcon(CharIconEngine("\uf4ad")), "View metadata", self.show_entity_metadata
-        )
-        self._menu.addSeparator()
         self._edit_action = self._menu.addAction(self._cube_pen_icon, "Edit...", self.edit_selected)
         self._remove_action = self._menu.addAction(self._cube_minus_icon, "Remove...", self.remove_selected)
         self._menu.addSeparator()
@@ -104,8 +105,6 @@ class EntityTreeView(CopyTreeView):
 
     def connect_signals(self):
         """Connects signals."""
-        self.expanded.connect(self._resize_first_column_to_contents)
-        self.collapsed.connect(self._resize_first_column_to_contents)
         self.selectionModel().selectionChanged.connect(self._handle_selection_changed)
         self._menu.aboutToShow.connect(self._spine_db_editor.refresh_copy_paste_actions)
 
@@ -127,10 +126,8 @@ class EntityTreeView(CopyTreeView):
     @Slot()
     def _fetch_more_visible(self):
         model = self.model()
-        for item in model.visit_all():
+        for item in model.visit_all(view=self):
             index = model.index_from_item(item)
-            if not self.isExpanded(index):
-                continue
             last = model.index(model.rowCount(index) - 1, 0, index)
             if self.visualRect(last).intersects(self.viewport().rect()) and model.canFetchMore(index):
                 model.fetchMore(index)
@@ -139,10 +136,6 @@ class EntityTreeView(CopyTreeView):
         super().verticalScrollbarValueChanged(value)
         self._fetch_more_timer.start()
 
-    @Slot(QModelIndex)
-    def _resize_first_column_to_contents(self, _index=None):
-        self.resizeColumnToContents(0)
-
     @Slot(QItemSelection, QItemSelection)
     def _handle_selection_changed(self, selected, deselected):
         """Classifies selection by item type and emits signal."""
@@ -170,26 +163,20 @@ class EntityTreeView(CopyTreeView):
     @busy_effect
     def fully_expand(self):
         """Expands selected indexes and all their children."""
-        self.expanded.disconnect(self._resize_first_column_to_contents)
         model = self.model()
         indexes = [index for index in self.selectionModel().selectedIndexes() if index.column() == 0]
         for index in indexes:
             for item in model.visit_all(index):
                 self.expand(model.index_from_item(item))
-        self.expanded.connect(self._resize_first_column_to_contents)
-        self._resize_first_column_to_contents()
 
     @busy_effect
     def fully_collapse(self):
         """Collapses selected indexes and all their children."""
-        self.collapsed.disconnect(self._resize_first_column_to_contents)
         model = self.model()
         indexes = [index for index in self.selectionModel().selectedIndexes() if index.column() == 0]
         for index in indexes:
             for item in model.visit_all(index):
                 self.collapse(model.index_from_item(item))
-        self.collapsed.connect(self._resize_first_column_to_contents)
-        self._resize_first_column_to_contents()
 
     def export_selected(self):
         """Exports data from selected indexes using the connected Spine db editor."""
@@ -197,22 +184,11 @@ class EntityTreeView(CopyTreeView):
 
     def remove_selected(self):
         """Removes selected indexes using the connected Spine db editor."""
-        self._spine_db_editor.show_remove_entity_tree_items_form(self._selected_indexes)
+        self._spine_db_editor.remove_entity_tree_items(self._selected_indexes)
 
     def manage_relationships(self):
         self._spine_db_editor.show_manage_relationships_form(self._context_item)
 
-    def show_entity_metadata(self):
-        """Shows entity's metadata."""
-        db_map_ids = {}
-        for index in set(self._selected_indexes.get("object", {})) | set(
-            self._selected_indexes.get("relationship", {})
-        ):
-            item = self.model().item_from_index(index)
-            for db_map, id_ in item.db_map_ids.items():
-                db_map_ids.setdefault(db_map, list()).append(id_)
-        self._spine_db_editor.show_db_map_entity_metadata(db_map_ids)
-
     def contextMenuEvent(self, event):
         """Shows context menu.
 
@@ -271,9 +247,8 @@ class EntityTreeView(CopyTreeView):
         item_has_children = item.has_children()
         self._fully_expand_action.setEnabled(item_has_children)
         self._fully_collapse_action.setEnabled(item_has_children)
-        self._add_relationships_action.setEnabled(item.item_type in ("root", "relationship_class"))
+        self._add_relationships_action.setEnabled(item.item_type == "relationship_class")
         self._manage_relationships_action.setEnabled(item.item_type in ("root", "relationship_class"))
-        self._show_entity_metadata_action.setEnabled(item.item_type in ("object", "relationship"))
         read_only = item.item_type in ("root", "members")
         self._export_action.setEnabled(not read_only)
         self._edit_action.setEnabled(not read_only)
@@ -365,8 +340,8 @@ class ObjectTreeView(EntityTreeView):
 
     def duplicate_object(self):
         """Duplicates the object at the current index using the connected Spine db editor."""
-        index = self.currentIndex()
-        self._spine_db_editor.duplicate_object(index)
+        object_item = self.currentIndex().internalPointer()
+        self._spine_db_editor.duplicate_object(object_item)
 
     def add_object_group(self):
         index = self.currentIndex()
@@ -408,14 +383,8 @@ class ItemTreeView(CopyTreeView):
 
     def connect_signals(self):
         """Connects signals."""
-        self.expanded.connect(self._resize_first_column_to_contents)
-        self.collapsed.connect(self._resize_first_column_to_contents)
         self._menu.aboutToShow.connect(self._spine_db_editor.refresh_copy_paste_actions)
 
-    @Slot(QModelIndex)
-    def _resize_first_column_to_contents(self, _index=None):
-        self.resizeColumnToContents(0)
-
     def remove_selected(self):
         """Removes items selected in the view."""
         raise NotImplementedError()
@@ -484,6 +453,8 @@ class ToolFeatureTreeView(ItemTreeView):
                         db_map_typed_data_to_rm[db_item.db_map]["tool_feature"].add(tool_feat_item.id)
                         continue
                     tool_feat_meth_root_item = tool_feat_item.child(1)
+                    if tool_feat_meth_root_item is None:
+                        continue
                     for tool_feat_meth_item in reversed(tool_feat_meth_root_item.children):
                         if tool_feat_meth_item in items:
                             db_map_typed_data_to_rm[db_item.db_map]["tool_feature_method"].add(tool_feat_meth_item.id)
@@ -515,8 +486,13 @@ class AlternativeScenarioTreeView(ItemTreeView):
         """Initialize the view."""
         super().__init__(parent=parent)
         self._selected_alternative_ids = dict()
+        self._generate_scenarios_action = None
         self.setMouseTracking(True)
 
+    def reset(self):
+        super().reset()
+        self._selected_alternative_ids = dict()
+
     def connect_signals(self):
         """Connects signals."""
         super().connect_signals()
@@ -529,14 +505,22 @@ class AlternativeScenarioTreeView(ItemTreeView):
         delegate.data_committed.connect(self.model().setData)
         self.setItemDelegateForColumn(0, delegate)
 
+    def populate_context_menu(self):
+        """See base class."""
+        self._generate_scenarios_action = self._menu.addAction("Generate scenarios...", self._open_scenario_generator)
+        self._menu.addSeparator()
+        super().populate_context_menu()
+
     def _db_map_alt_ids_from_selection(self, selection):
         db_map_ids = {}
         for index in selection.indexes():
             if index.column() != 0:
                 continue
             item = self.model().item_from_index(index)
-            if item.item_type == "alternative" and item.id:
+            if item.item_type == "alternative" and hasattr(item, "id") and item.id:
                 db_map_ids.setdefault(item.db_map, set()).add(item.id)
+            if item.item_type == "scenario_alternative" and hasattr(item, "alternative_id") and item.alternative_id:
+                db_map_ids.setdefault(item.db_map, set()).add(item.alternative_id)
         return db_map_ids
 
     def _db_map_scen_alt_ids_from_selection(self, selection):
@@ -545,7 +529,7 @@ class AlternativeScenarioTreeView(ItemTreeView):
             if index.column() != 0:
                 continue
             item = self.model().item_from_index(index)
-            if item.item_type == "scenario_alternative root":
+            if item.item_type == "scenario_alternative" and hasattr(item, "alternative_id_list"):
                 db_map_ids.setdefault(item.db_map, set()).update(item.alternative_id_list)
         return db_map_ids
 
@@ -585,9 +569,13 @@ class AlternativeScenarioTreeView(ItemTreeView):
                     db_map_typed_data_to_rm[db_item.db_map]["scenario"].add(scen_item.id)
                     continue
                 scen_alt_root_item = scen_item.scenario_alternative_root_item
+                if not scen_alt_root_item.non_empty_children:
+                    continue
                 curr_alt_id_list = scen_alt_root_item.alternative_id_list
                 new_alt_id_list = [
-                    id_ for alt_item, id_ in zip(scen_alt_root_item.children, curr_alt_id_list) if alt_item not in items
+                    id_
+                    for alt_item, id_ in zip(scen_alt_root_item.non_empty_children, curr_alt_id_list)
+                    if alt_item not in items
                 ]
                 if new_alt_id_list != curr_alt_id_list:
                     item = {"id": scen_item.id, "alternative_id_list": ",".join([str(id_) for id_ in new_alt_id_list])}
@@ -598,6 +586,9 @@ class AlternativeScenarioTreeView(ItemTreeView):
 
     def update_actions_availability(self, item):
         """See base class."""
+        self._generate_scenarios_action.setEnabled(
+            isinstance(item, AlternativeLeafItem) and bool(self._selected_alternative_ids.get(item.db_map))
+        )
 
     def dragMoveEvent(self, event):
         super().dragMoveEvent(event)
@@ -611,10 +602,24 @@ class AlternativeScenarioTreeView(ItemTreeView):
         if event.source() is self:
             event.accept()
 
+    def _open_scenario_generator(self):
+        """Opens the scenario generator dialog."""
+        item = self.model().item_from_index(self.currentIndex())
+        if not isinstance(item, AlternativeLeafItem):
+            return
+        included_ids = set()
+        alternatives = list()
+        db_map = item.db_map
+        for id_ in self._selected_alternative_ids.get(db_map, ()):
+            if id_ not in included_ids:
+                alternatives.append(self._spine_db_editor.db_mngr.get_item(db_map, "alternative", id_))
+                included_ids.add(id_)
+        generator = ScenarioGenerator(self, db_map, alternatives, self._spine_db_editor)
+        generator.show()
+
 
 class ParameterValueListTreeView(ItemTreeView):
-    """Custom QTreeView class for parameter_value_list in SpineDBEditor.
-    """
+    """Custom QTreeView class for parameter_value_list in SpineDBEditor."""
 
     def __init__(self, parent):
         """Initialize the view."""
@@ -637,7 +642,7 @@ class ParameterValueListTreeView(ItemTreeView):
 
     def update_actions_availability(self, item):
         """See base class."""
-        self._open_in_editor_action.setEnabled(item.item_type == "value")
+        self._open_in_editor_action.setEnabled(item.item_type == "list_value")
 
     def open_in_editor(self):
         """Opens the parameter_value editor for the first selected cell."""
@@ -649,36 +654,17 @@ class ParameterValueListTreeView(ItemTreeView):
         if not self.selectionModel().hasSelection():
             return
         db_map_typed_data_to_rm = {}
-        db_map_data_to_upd = {}
         items = [self.model().item_from_index(index) for index in self.selectionModel().selectedIndexes()]
         for db_item in self.model()._invisible_root_item.children:
-            db_map_typed_data_to_rm[db_item.db_map] = {"parameter_value_list": set()}
-            db_map_data_to_upd[db_item.db_map] = []
-            for list_item in reversed(db_item.children[:-1]):
-                if list_item.id:
-                    if list_item in items:
-                        db_map_typed_data_to_rm[db_item.db_map]["parameter_value_list"].add(list_item.id)
-                        continue
-                    curr_value_list = list_item.value_list
-                    new_value_list = [
-                        value
-                        for value_item, value in zip(list_item.children, curr_value_list)
-                        if value_item not in items
-                    ]
-                    if not new_value_list:
-                        db_map_typed_data_to_rm[db_item.db_map]["parameter_value_list"].add(list_item.id)
-                        continue
-                    if new_value_list != curr_value_list:
-                        item = {"id": list_item.id, "value_list": [bytes(val, "UTF-8") for val in new_value_list]}
-                        db_map_data_to_upd[db_item.db_map].append(item)
+            db_map_typed_data_to_rm[db_item.db_map] = {"parameter_value_list": set(), "list_value": set()}
+            for list_item in db_item.children[:-1]:
+                if list_item.id is None:
+                    continue
+                if list_item in items:
+                    db_map_typed_data_to_rm[db_item.db_map]["parameter_value_list"].add(list_item.id)
+                    removed_value_item_ids = {x.id for x in list_item.children[:-1]}
                 else:
-                    # WIP lists, just remove everything selected
-                    if list_item in items:
-                        db_item.remove_children(list_item.child_number(), 1)
-                        continue
-                    for value_item in reversed(list_item.children[:-1]):
-                        if value_item in items:
-                            list_item.remove_children(value_item.child_number(), 1)
-        self.model().db_mngr.update_parameter_value_lists(db_map_data_to_upd)
+                    removed_value_item_ids = {x.id for x in list_item.children[:-1] if x in items}
+                db_map_typed_data_to_rm[db_item.db_map]["list_value"].update(removed_value_item_ids)
         self.model().db_mngr.remove_items(db_map_typed_data_to_rm)
         self.selectionModel().clearSelection()
