@@ -43,11 +43,10 @@ class SpineDBParcel:
         return self._data
 
     def _get_fields(self, db_map, item_type, field, ids):
-        # FIXME: We need to fetch all the items here
         if ids is Asterisk:
-            fields = {x.get(field) for x in self.db_mngr.get_items(db_map, item_type)}
+            fields = {x.get(field) for x in self.db_mngr.get_items(db_map, item_type, only_visible=False)}
         else:
-            fields = {self.db_mngr.get_field(db_map, item_type, id_, field) for id_ in ids}
+            fields = {self.db_mngr.get_field(db_map, item_type, id_, field, only_visible=False) for id_ in ids}
         fields.discard(None)
         return fields
 
@@ -187,15 +186,43 @@ class SpineDBParcel:
             }
         )
 
+    def push_feature_ids(self, db_map_ids):
+        """Pushes feature ids."""
+        self._update_ids(db_map_ids, "feature_ids")
+
+    def push_tool_ids(self, db_map_ids):
+        """Pushes tool ids."""
+        self._update_ids(db_map_ids, "tool_ids")
+
+    def push_tool_feature_ids(self, db_map_ids):
+        """Pushes tool_feature ids."""
+        self._update_ids(db_map_ids, "tool_feature_ids")
+        self.push_feature_ids(
+            {db_map: self._get_fields(db_map, "tool_feature", "feature_id", ids) for db_map, ids in db_map_ids.items()}
+        )
+        self.push_tool_ids(
+            {db_map: self._get_fields(db_map, "tool_feature", "tool_id", ids) for db_map, ids in db_map_ids.items()}
+        )
+
+    def push_tool_feature_method_ids(self, db_map_ids):
+        """Pushes tool_feature_method ids."""
+        self._update_ids(db_map_ids, "tool_feature_method_ids")
+        self.push_tool_feature_ids(
+            {
+                db_map: self._get_fields(db_map, "tool_feature_method", "tool_feature_id", ids)
+                for db_map, ids in db_map_ids.items()
+            }
+        )
+
     def full_push_object_class_ids(self, db_map_ids):
         """Pushes parameter definitions associated with given object classes.
         This essentially full_pushes the object classes and their parameter definitions.
         """
         param_def_ids = self.db_mngr.db_map_ids(
-            self.db_mngr.find_cascading_parameter_data(db_map_ids, "parameter_definition")
+            self.db_mngr.find_cascading_parameter_data(db_map_ids, "parameter_definition", only_visible=False)
         )
         self.push_parameter_definition_ids(param_def_ids, "object")
-        db_map_ids = {db_map: ids for db_map, ids in db_map_ids.items() if not param_def_ids[db_map]}
+        db_map_ids = {db_map: ids for db_map, ids in db_map_ids.items() if db_map not in param_def_ids}
         self.push_object_class_ids(db_map_ids)
 
     def full_push_relationship_class_ids(self, db_map_ids):
@@ -203,10 +230,10 @@ class SpineDBParcel:
         This essentially full_pushes the relationships classes, their parameter definitions, and their member object classes.
         """
         param_def_ids = self.db_mngr.db_map_ids(
-            self.db_mngr.find_cascading_parameter_data(db_map_ids, "parameter_definition")
+            self.db_mngr.find_cascading_parameter_data(db_map_ids, "parameter_definition", only_visible=False)
         )
         self.push_parameter_definition_ids(param_def_ids, "relationship")
-        db_map_ids = {db_map: ids for db_map, ids in db_map_ids.items() if not param_def_ids[db_map]}
+        db_map_ids = {db_map: ids for db_map, ids in db_map_ids.items() if db_map not in param_def_ids}
         self.push_relationship_class_ids(db_map_ids)
 
     def full_push_object_ids(self, db_map_ids):
@@ -214,10 +241,14 @@ class SpineDBParcel:
         This essentially full_pushes objects, their relationships, all the parameter values, and all the necessary classes,
         definitions, and lists.
         """
-        self.full_push_relationship_ids(self.db_mngr.db_map_ids(self.db_mngr.find_cascading_relationships(db_map_ids)))
-        param_val_ids = self.db_mngr.db_map_ids(self.db_mngr.find_cascading_parameter_values_by_entity(db_map_ids))
+        self.full_push_relationship_ids(
+            self.db_mngr.db_map_ids(self.db_mngr.find_cascading_relationships(db_map_ids, only_visible=False))
+        )
+        param_val_ids = self.db_mngr.db_map_ids(
+            self.db_mngr.find_cascading_parameter_values_by_entity(db_map_ids, only_visible=False)
+        )
         self.push_parameter_value_ids(param_val_ids, "object")
-        db_map_ids = {db_map: ids for db_map, ids in db_map_ids.items() if not param_val_ids[db_map]}
+        db_map_ids = {db_map: ids for db_map, ids in db_map_ids.items() if db_map not in param_val_ids}
         self.push_object_ids(db_map_ids)
         self.push_object_group_ids(self.db_mngr.db_map_ids(self.db_mngr.find_groups_by_entity(db_map_ids)))
 
@@ -226,9 +257,11 @@ class SpineDBParcel:
         This essentially full_pushes relationships, their parameter values, and all the necessary classes,
         definitions, and lists.
         """
-        param_val_ids = self.db_mngr.db_map_ids(self.db_mngr.find_cascading_parameter_values_by_entity(db_map_ids))
+        param_val_ids = self.db_mngr.db_map_ids(
+            self.db_mngr.find_cascading_parameter_values_by_entity(db_map_ids, only_visible=False)
+        )
         self.push_parameter_value_ids(param_val_ids, "relationship")
-        db_map_ids = {db_map: ids for db_map, ids in db_map_ids.items() if not param_val_ids[db_map]}
+        db_map_ids = {db_map: ids for db_map, ids in db_map_ids.items() if db_map not in param_val_ids}
         self.push_relationship_ids(db_map_ids)
 
     def inner_push_object_ids(self, db_map_ids):
@@ -238,9 +271,14 @@ class SpineDBParcel:
         """
         for db_map, ids in db_map_ids.items():
             self._setdefault(db_map)["object_ids"].update(ids)
-        self.inner_push_relationship_ids(self.db_mngr.db_map_ids(self.db_mngr.find_cascading_relationships(db_map_ids)))
+        self.inner_push_relationship_ids(
+            self.db_mngr.db_map_ids(self.db_mngr.find_cascading_relationships(db_map_ids, only_visible=False))
+        )
         self.inner_push_parameter_value_ids(
-            self.db_mngr.db_map_ids(self.db_mngr.find_cascading_parameter_values_by_entity(db_map_ids)), "object"
+            self.db_mngr.db_map_ids(
+                self.db_mngr.find_cascading_parameter_values_by_entity(db_map_ids, only_visible=False)
+            ),
+            "object",
         )
 
     def inner_push_relationship_ids(self, db_map_ids):
@@ -249,7 +287,10 @@ class SpineDBParcel:
         for db_map, ids in db_map_ids.items():
             self._setdefault(db_map)["relationship_ids"].update(ids)
         self.inner_push_parameter_value_ids(
-            self.db_mngr.db_map_ids(self.db_mngr.find_cascading_parameter_values_by_entity(db_map_ids)), "relationship"
+            self.db_mngr.db_map_ids(
+                self.db_mngr.find_cascading_parameter_values_by_entity(db_map_ids, only_visible=False)
+            ),
+            "relationship",
         )
 
     def inner_push_parameter_value_ids(self, db_map_ids, entity_type):
@@ -257,18 +298,18 @@ class SpineDBParcel:
         for db_map, ids in db_map_ids.items():
             self._setdefault(db_map)[entity_type + "_parameter_value_ids"].update(ids)
 
-    def _update_ids(self, db_map_ids, which):
+    def _update_ids(self, db_map_ids, key):
         """Updates ids for given database item.
 
         Args:
             db_map_ids (dict): mapping from :class:`DatabaseMappingBase` to ids or ``Asterisk``
-            which (str): item name
+            key (str): the key
         """
         for db_map, ids in db_map_ids.items():
             if ids is Asterisk:
-                self._setdefault(db_map)[which] = ids
+                self._setdefault(db_map)[key] = ids
             else:
-                current = self._setdefault(db_map)[which]
+                current = self._setdefault(db_map)[key]
                 if current is not Asterisk:
                     current.update(ids)
 
@@ -296,5 +337,9 @@ class SpineDBParcel:
             "alternative_ids": set(),
             "scenario_ids": set(),
             "scenario_alternative_ids": set(),
+            "feature_ids": set(),
+            "tool_ids": set(),
+            "tool_feature_ids": set(),
+            "tool_feature_method_ids": set(),
         }
         return self._data.setdefault(db_map, d)
